/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "SubscriptionsPanel.h"

#include <Components/FilterSelector/FilterSelector.h>

#include <QBoxLayout>
#include <QDebug>
#include <QListWidget>
#include <QPushButton>
#include <functional>

using namespace std;

SubscriptionsPanel::SubscriptionsPanel() : QWidget(nullptr)
{
    resize({500, 200});

    setWindowTitle("Subscriptions panel");
    setupUi();
}

SubscriptionsPanel::SubscriptionsPanel(const QList<Filter>& filters)
    : SubscriptionsPanel(filters.toStdList())
{
}

SubscriptionsPanel::SubscriptionsPanel(const std::list<Filter>& filters)
    : SubscriptionsPanel()
{
    setFilters(filters);
}

void SubscriptionsPanel::setFilters(const std::list<Filter>& filters)
{
    filtersList->clear();
    for (auto filter : filters)
        addFilter(filter);
}

void SubscriptionsPanel::setFilters(const QList<Filter>& filters)
{
    setFilters(filters.toStdList());
}

void SubscriptionsPanel::addFilter(const Filter& filter)
{
    filtersList->addItem(new QListWidgetItem(filter.toPrettyString()));
    emit filterAdded(filter);
}

void SubscriptionsPanel::setupUi()
{
    QVBoxLayout* layout = new QVBoxLayout;

    filtersList = std::make_unique<QListWidget>();
    filtersList->setSelectionBehavior(QAbstractItemView::SelectItems);
    filtersList->setSelectionMode(QAbstractItemView::SingleSelection);
    layout->addWidget(filtersList.get());

    QHBoxLayout* buttonsLayout = new QHBoxLayout;

    QPushButton* addButton = new QPushButton;
    addButton->setText("Add filter");
    buttonsLayout->addWidget(addButton);
    QPushButton* editButton = new QPushButton;
    editButton->setText("Edit filter");
    buttonsLayout->addWidget(editButton);
    QPushButton* removeButton = new QPushButton;
    removeButton->setText("Remove filter");
    buttonsLayout->addWidget(removeButton);

    layout->addLayout(buttonsLayout);

    setLayout(layout);

    connect(addButton, &QPushButton::clicked, this,
            [&]()
            {
                FilterSelector::selectFilter([&](const Filter& filter)
                                             { addFilter(filter); });
            });
    connect(editButton, &QPushButton::clicked, this,
            [&]()
            {
                if (filtersList->selectedItems().size() > 0)
                {
                    auto item = filtersList->selectedItems().at(0);

                    FilterSelector::selectFilter(
                        Filter::fromString(item->text()),
                        [=](const Filter& newFilter)
                        {
                            if (Filter::fromString(item->text()) != newFilter)
                            {
                                removeSubscription(item);
                                addFilter(newFilter);
                            }
                        });
                }
            });
    connect(removeButton, &QPushButton::clicked, this,
            [&]()
            {
                for (auto item : filtersList->selectedItems())
                    removeSubscription(item);
            });
}

void SubscriptionsPanel::removeSubscription(QListWidgetItem* item)
{
    emit filterRemoved(Filter::fromString(item->text()));
    delete item;
}
