/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "SkywardHubCore.h"

#include <Components/SaveConfigurationDialog/SaveConfigurationDialog.h>
#include <Core/MessageBroker/MessageBroker.h>
#include <Core/Window/Window.h>
#include <Modules/Module.h>
#include <Modules/ModulesList.h>
#include <Modules/SkywardHubStrings.h>

#include <QCloseEvent>
#include <QDir>
#include <QMessageBox>

SkywardHubCore& SkywardHubCore::getInstance()
{
    static SkywardHubCore instance;
    return instance;
}

void SkywardHubCore::init()
{

    QSettings userSettings("Skyward ER", "SkywardHub");

    userSettings.beginGroup("Customization");
    QString themePath =
        userSettings
            .value("Theme-Path", SkywardHubStrings::defaultThemeFilePath)
            .toString();

    activeTheme = userSettings.value("Theme-Name", "Default").toString();

    userSettings.endGroup();

    loadStyleSheet(themePath);

    XmlObject settings;
    settings.loadFromFile(SkywardHubStrings::defaultSettingsFilePath);

    QString settingsName =
        settings.getChildObjectValue(SkywardHubStrings::skywardHubInitFileTag);

    if (!settingsName.isEmpty())
    {
        XmlObject configuration;
        if (configuration.loadFromFile(SkywardHubStrings::defaultPrefabsFolder +
                                       settingsName))
        {

            loadConfigurationFromXmlObject(configuration);
            return;
        }
    }

    addNewWindow();
    saveConfiguration(SkywardHubStrings::defaultConfigurationFileName,
                      SkywardHubStrings::defaultConfigurationIconPath, "");
}

SkywardHubCore::SkywardHubCore()
{
    checkDefaultFolders();
    buildHubMenuActions();
}

void SkywardHubCore::checkDefaultFolders()
{
    /*
     * If SkywardHubConfig is missing, it create all the missing folder
     */

    if (!QDir(SkywardHubStrings::defaultConfigurationFolder).exists())
    {
        QDir().mkdir(SkywardHubStrings::defaultConfigurationFolder);
    }
}

void SkywardHubCore::buildHubMenuActions()
{
    QAction* save = new QAction("Save configuration");
    hubMenuActionsList.append(save);
    connect(save, &QAction::triggered, this,
            [this]() { saveConfigurationPrompt(); });

    QAction* newWindow = new QAction("New Window");
    hubMenuActionsList.append(newWindow);
    connect(newWindow, &QAction::triggered, this,
            &SkywardHubCore::addNewWindow);

    buildThemesMenuSelector();
}

void SkywardHubCore::loadConfigurationFromXmlObject(XmlObject configuration)
{
    XmlObject windows;
    windows = configuration.childAt(0);

    for (int i = 0; i < windows.childCount(); i++)
    {
        addWindowFromXml(windows.childAt(i));
    }
}

void SkywardHubCore::addNewWindow()
{
    std::shared_ptr<Module> empty =
        ModulesList::getInstance().instantiateModule(ModuleId::EMPTY);
    return addWindowFromXml(empty->toXmlObject());
}

void SkywardHubCore::setupMenu(QMenu& menu)
{
    menu.addActions(hubMenuActionsList);
    menu.addMenu(themeSelectionMenu);
}

void SkywardHubCore::addWindowFromXml(XmlObject settings)
{
    Window* newWindow = new Window(settings);
    newWindow->show();
    newWindow->activateWindow();
    newWindow->setStyleSheet(activeThemeStyleSheet);
    windowsList.push_back(newWindow);

    connect(newWindow, &Window::onWindowClosed, this,
            [this, newWindow]()
            {
                windowsList.removeAt(windowsList.indexOf(newWindow));
                delete newWindow;
            });
}

void SkywardHubCore::saveConfigurationPrompt()
{
    // Ask configuration data to the user
    std::unique_ptr<SaveConfigurationDialog> saveDialog =
        std::make_unique<SaveConfigurationDialog>();
    saveDialog->setConfigName(SkywardHubStrings::defaultConfigurationFileName);
    saveDialog->setConfigIconPath(
        SkywardHubStrings::defaultConfigurationIconPath);
    QString configName;
    QString configIconPath;
    QString description;
    saveDialog->pickConfigData(&configName, &configIconPath, &description);

    QString resultTxt;
    resultTxt = saveConfiguration(configName, configIconPath, description);

    QMessageBox resultMsg;
    resultMsg.setText(resultTxt);
    resultMsg.exec();
}

QString SkywardHubCore::saveConfiguration(QString configName,
                                          QString configIconPath,
                                          QString description)
{

    XmlObject configuration(SkywardHubStrings::configurationNameValue);
    configuration.addAttribute(
        SkywardHubStrings::configurationIconPathAttribute, configIconPath);
    configuration.addAttribute(
        SkywardHubStrings::configurationDescriptionAttribute, description);

    QString resultTxt = "Impossible to save";

    if (!configName.isEmpty())
    {
        configName = configName.replace(".xml", "");
        configuration.addAttribute(
            SkywardHubStrings::configurationNameAttribute, configName);
        if (configName.indexOf(".xml") < 0)
        {
            configName.append(".xml");
        }

        if (!QDir(SkywardHubStrings::defaultPrefabsFolder).exists())
        {
            QDir().mkpath(SkywardHubStrings::defaultPrefabsFolder);
        }

        XmlObject windows("windows");

        // Retrieve the setting of each window and save them in the file
        for (int i = 0; i < windowsList.size(); i++)
        {
            windows.addChild(windowsList[i]->toXmlObject());
        }

        configuration.addChild(windows);

        if (configuration.writeToFile(SkywardHubStrings::defaultPrefabsFolder +
                                      configName))
        {
            resultTxt = "The configuration has been saved in " + configName;
            XmlObject settings;
            settings.setObjectName(SkywardHubStrings::settingsObjectName);
            XmlObject defaultConfig(SkywardHubStrings::skywardHubInitFileTag);
            defaultConfig.setTextValue(configName);
            settings.addChild(defaultConfig);
            settings.writeToFile(SkywardHubStrings::defaultSettingsFilePath);
        }
    }

    return resultTxt;
}

void SkywardHubCore::buildThemesMenuSelector()
{
    themeSelectionMenu = new QMenu("Themes");

    for (auto pair : SkywardHubStrings::defaultThemes.toStdMap())
    {
        QAction* builtinThemeAction = new QAction(pair.first);
        builtinThemeAction->setData(pair.second);
        themeSelectionMenu->addAction(builtinThemeAction);
    }

    themeSelectionMenu->addSeparator();

    QList<QAction*> availableThemes = fetchAvailableThemes();

    themeSelectionMenu->addActions(availableThemes);

    if (availableThemes.size() > 0)
    {
        themeSelectionMenu->addSeparator();
    }

    QAction* openThemesFolder = new QAction("Open Folder");

    themeSelectionMenu->addAction(openThemesFolder);

    initializeActiveTheme();

    connect(themeSelectionMenu, &QMenu::triggered, this,
            &SkywardHubCore::onThemeSelected);
}

QList<QAction*> SkywardHubCore::fetchAvailableThemes()
{
    std::unique_ptr<QDir> themesDirectory =
        std::make_unique<QDir>(SkywardHubStrings::defaultThemesFolderPath);

    QStringList entries = themesDirectory->entryList(QDir::Files, QDir::Name);

    QList<QAction*> actions;

    for (QString& fileName : entries)
    {
        QAction* newAction = new QAction(fileName.remove(".qss"));
        newAction->setData(SkywardHubStrings::defaultThemesFolderPath + "/" +
                           fileName + ".qss");
        actions.append(newAction);
    }

    return actions;
}

void SkywardHubCore::onThemeSelected(QAction* action)
{

    if (action->text() == "Open Folder")
    {
        QDesktopServices::openUrl(
            QUrl::fromLocalFile(SkywardHubStrings::defaultConfigurationFolder));
        return;
    }

    QString themePath = action->data().toString();
    loadStyleSheet(themePath);

    unsetOldActiveTheme();
    setActiveTheme(action);
}

void SkywardHubCore::loadStyleSheet(QString path)
{
    QFile styleSheetFile(path);
    styleSheetFile.open(QFile::ReadOnly);
    activeThemeStyleSheet = styleSheetFile.readAll();

    for (Window* window : windowsList)
    {
        window->setStyleSheet(activeThemeStyleSheet);
    }

    qApp->setStyleSheet(activeThemeStyleSheet);

    styleSheetFile.close();
}

void SkywardHubCore::setActiveTheme(QAction* action)
{

    action->setEnabled(false);

    QSettings settings("Skyward ER", "SkywardHub");
    settings.beginGroup("Customization");
    settings.setValue("Theme-Name", action->text());
    settings.setValue("Theme-Path", action->data().toString());
    settings.endGroup();

    activeTheme = action->text();
}

void SkywardHubCore::unsetOldActiveTheme()
{
    QAction* lastActiveTheme = getActiveThemeAction();
    if (lastActiveTheme == nullptr)
        return;

    lastActiveTheme->setEnabled(true);
}

void SkywardHubCore::initializeActiveTheme()
{
    QAction* currentActiveTheme = getActiveThemeAction();
    if (currentActiveTheme == nullptr)
        return;

    currentActiveTheme->setEnabled(false);
}

QAction* SkywardHubCore::getActiveThemeAction()
{
    if (activeTheme == "")
        return nullptr;

    const QList<QAction*> menuActions = themeSelectionMenu->actions();

    auto foundAction = std::find_if(menuActions.begin(), menuActions.end(),
                                    [this](QAction* act)
                                    { return act->text() == activeTheme; });

    if (foundAction != menuActions.end())
    {
        return *foundAction;
    }

    return nullptr;
}
