#include "FileStreamModule.h"

#include <Core/MessageBroker/MessageBroker.h>

#include <QDebug>
#include <QDesktopServices>
#include <QDir>
#include <QMessageBox>
#include <QUrl>

#include "Modules/SkywardHubStrings.h"
#include "ui_FileStreamModule.h"

FileStreamModule::FileStreamModule()
    : Module(ModuleId::FILE_STREAM), ui(new Ui::FileStreamModule)
{
    ui->setupUi(this);
    connectUI();

    ui->filePath_lineEdit->setText(SkywardHubStrings::defaultStreamFile);
}

FileStreamModule::~FileStreamModule()
{
    if (file.isOpen())
        file.close();
    deleteAllTopicViews();
    delete ui;
}

XmlObject FileStreamModule::toXmlObject()
{
    XmlObject obj = Module::toXmlObject();

    if (getFilePath() != "")
        obj.addAttribute("FilePath", getFilePath());

    for (auto& topic : topicViewsList)
    {
        XmlObject subscription("Subscription");
        subscription.setTextValue(topic.first->text().trimmed());
        obj.addChild(subscription);
    }
    return obj;
}

void FileStreamModule::fromXmlObject(const XmlObject& xmlObject)
{
    if (xmlObject.hasAttribute("FilePath"))
    {
        ui->filePath_lineEdit->setText(xmlObject.getAttribute("FilePath"));
    }

    for (int i = 0; i < xmlObject.childCount(); i++)
    {
        XmlObject child = xmlObject.childAt(i);
        if (child.getObjectName() == "Subscription" &&
            child.getTextValue() != "")
        {
            addTopic(child.getTextValue());
        }
    }
}

void FileStreamModule::connectUI()
{
    connect(ui->start_button, &QPushButton::clicked, this,
            &FileStreamModule::onStartClicked);
    connect(ui->stop_button, &QPushButton::clicked, this,
            &FileStreamModule::onStopClicked);
    connect(ui->addTopic_button, &QPushButton::clicked, this,
            &FileStreamModule::onAddTopicClicked);
    connect(ui->showFile_button, &QPushButton::clicked, this,
            &FileStreamModule::onShowFileClick);
}

QString FileStreamModule::getFilePath() const
{
    return ui->filePath_lineEdit->text().trimmed();
}

QString FileStreamModule::getTopicName() const
{
    return ui->topicName_lineEdit->text().trimmed();
}

void FileStreamModule::resetTopicName() { ui->topicName_lineEdit->clear(); }

void FileStreamModule::onMsgReceived(const Message& msg)
{
    //    if(file.isOpen() ){
    //        QTextStream out(&file);
    //        out << msg.payload() << Qt::endl;
    //    }

    QFile file(getFilePath());
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
        return;

    QTextStream out(&file);
    out << msg.toString() << Qt::endl;
}

void FileStreamModule::onAddTopicClicked()
{
    QString topicName = getTopicName();
    resetTopicName();
    if (topicName != "")
    {
        addTopic(topicName);
    }
}

void FileStreamModule::onStartClicked()
{
    file.setFileName(getFilePath());
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
    {  // Append QIODevice::ReadWrite
        error(tr("FileStream"),
              tr("Error occurred while opening file ") + getFilePath());
        return;
    }

    file.resize(0);
    disableOnStartElement();
    for (auto& topic : topicViewsList)
    {
        topic.first->setEnabled(false);
        topic.second = MessageBroker::getInstance().subscribe(
            Filter::fromString(topic.first->text().trimmed()),
            [this](const Message& message, const Filter& filter)
            { onMsgReceived(message); });
    }
}

void FileStreamModule::onStopClicked()
{
    if (file.isOpen())
        file.close();
    enableElementOnStop();

    for (auto& topicPair : topicViewsList)
    {
        topicPair.first->setEnabled(true);
        topicPair.second = nullptr;
    }
}

void FileStreamModule::onTopicEdit()
{
    auto erasable = std::remove_if(
        topicViewsList.begin(), topicViewsList.end(),
        [this](std::pair<QLineEdit*, std::unique_ptr<Subscription>>& value)
        {
            if (value.first->text().trimmed() == "")
            {
                value.first->deleteLater();
                return true;
            };
            return false;
        });

    topicViewsList.erase(erasable, topicViewsList.end());
}

void FileStreamModule::disableOnStartElement()
{
    ui->start_button->setEnabled(false);
    ui->addTopic_button->setEnabled(false);
    ui->filePath_lineEdit->setEnabled(false);
}

void FileStreamModule::enableElementOnStop()
{
    ui->start_button->setEnabled(true);
    ui->addTopic_button->setEnabled(true);
    ui->filePath_lineEdit->setEnabled(true);
    ui->writeOnFile_radioButton->setEnabled(true);
}

void FileStreamModule::onShowFileClick()
{
    QDir dir(SkywardHubStrings::defaultLogsFolder);
    if (dir.exists())
    {
        QDesktopServices::openUrl(QUrl::fromLocalFile(getFilePath()));
    }
    else
    {
        QString msg = "The file does not exist." + getFilePath();
        QMessageBox msgBox;
        msgBox.setText(msg);
        msgBox.exec();
    }
}

void FileStreamModule::addTopic(const QString& topic)
{
    QLineEdit* topicView = createTopicView(topic);

    topicViewsList.push_back(
        std::pair(topicView, std::unique_ptr<Subscription>(nullptr)));
    ui->topics_layout->addWidget(topicView);
}

QLineEdit* FileStreamModule::createTopicView(const QString& topic) const
{
    QLineEdit* topicView = new QLineEdit();
    topicView->setText(topic);

    connect(topicView, &QLineEdit::textEdited, this,
            &FileStreamModule::onTopicEdit);
    return topicView;
}

void FileStreamModule::deleteAllTopicViews()
{
    for (auto& topic : topicViewsList)
    {
        ui->topics_layout->removeWidget(topic.first);
        topic.first->deleteLater();
    }
    topicViewsList.clear();
}
