/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "IncomingMessagesViewerModule.h"

#include <QHBoxLayout>
#include <QTime>

IncomingMessagesViewerModule::IncomingMessagesViewerModule()
    : Module(ModuleId::INCOMING_MESSAGES_VIEWER)
{
    setupUi();
    customContextMenuActionSetup();

    subscriptionsPanel = std::make_unique<SubscriptionsPanel>();
    subscriptionsPanel->setWindowTitle("Graph subscriptions");
    connect(subscriptionsPanel.get(), &SubscriptionsPanel::filterAdded, this,
            &IncomingMessagesViewerModule::onFilterAdded);
    connect(subscriptionsPanel.get(), &SubscriptionsPanel::filterRemoved, this,
            &IncomingMessagesViewerModule::onFilterRemoved);
}

IncomingMessagesViewerModule::~IncomingMessagesViewerModule() {}

XmlObject IncomingMessagesViewerModule::toXmlObject()
{
    XmlObject obj = Module::toXmlObject();

    obj.addAttribute("keepOnlyLastMessage", keepOnlyLastMessage ? "1" : "0");
    obj.addAttribute("useTimestamp", useTimestamp ? "1" : "0");

    for (auto& pair : filters)
    {
        XmlObject subscription("subscription");
        subscription.addAttribute("filter", pair.first.toString());
        obj.addChild(subscription);
    }

    return obj;
}

void IncomingMessagesViewerModule::fromXmlObject(const XmlObject& xmlObject)
{
    keepOnlyLastMessage = xmlObject.getAttribute("keepOnlyLastMessage") == "1";
    keepOnlyLastMessageAction->setChecked(keepOnlyLastMessage);
    useTimestamp = xmlObject.getAttribute("useTimestamp") == "1";
    useTimestampAction->setChecked(useTimestamp);

    for (int i = 0; i < xmlObject.childCount(); i++)
    {
        XmlObject child = xmlObject.childAt(i);
        if (child.getObjectName() == "subscription")
            onFilterAdded(Filter::fromString(child.getAttribute("filter")));
    }
}

void IncomingMessagesViewerModule::onSubscribeClicked()
{
    std::list<Filter> filterKeys;

    for (auto it = filters.begin(); it != filters.end(); it++)
    {
        filterKeys.push_back(it->first);
    }

    subscriptionsPanel->setFilters(filterKeys);
    subscriptionsPanel->show();
}

void IncomingMessagesViewerModule::onFilterAdded(const Filter& filter)
{
    std::unique_ptr<Subscription> subscription =
        MessageBroker::getInstance().subscribe(
            filter,
            [&](const Message& message, const Filter& filter)
            {
                QString oldText = "";
                QString time    = "";

                // Save the old text
                if (!keepOnlyLastMessage)
                    oldText = edit->toPlainText().mid(0, 10000);

                if (useTimestamp)
                {
                    // If available  show the timestamp
                    auto timestamp = message.getField("timestamp");
                    time += "[";
                    time += QString::number(
                        (float)timestamp.getUnsignedInteger() / 1e6, 'f', 1);
                    time += "] ";
                }
                else
                {
                    time +=
                        "[" + QTime::currentTime().toString("hh.mm.ss") + "] ";
                }

                // Updated the content
                edit->setText(time + message.toString() + "\n" + oldText);
            });

    filters[filter] = std::move(subscription);
}

void IncomingMessagesViewerModule::onFilterRemoved(const Filter& filter)
{
    for (auto iter = filters.begin(); iter != filters.end();)
    {
        if (iter->first == filter)
        {
            filters.erase(iter++);
        }
        else
        {
            iter++;
        }
    }
}

void IncomingMessagesViewerModule::setupUi()
{
    edit = std::make_unique<QTextEdit>();
    edit->setReadOnly(true);
    edit->setContextMenuPolicy(Qt::ContextMenuPolicy::NoContextMenu);

    QHBoxLayout* layout = new QHBoxLayout();
    layout->setContentsMargins(0, 0, 0, 0);
    layout->addWidget(edit.get());
    setLayout(layout);
}

void IncomingMessagesViewerModule::customContextMenuActionSetup()
{
    std::shared_ptr<QAction> clear = std::make_shared<QAction>("Clear");
    connect(clear.get(), &QAction::triggered, this,
            [this]() { edit->clear(); });
    customContextMenuActions.push_back(clear);

    std::shared_ptr<QAction> subscriptions =
        std::make_shared<QAction>("Manage subscriptions");
    connect(subscriptions.get(), &QAction::triggered, this,
            &IncomingMessagesViewerModule::onSubscribeClicked);
    customContextMenuActions.push_back(subscriptions);

    keepOnlyLastMessageAction =
        std::make_shared<QAction>("Keep only last message");
    keepOnlyLastMessageAction->setCheckable(true);
    connect(keepOnlyLastMessageAction.get(), &QAction::toggled, this,
            [=](bool value) { keepOnlyLastMessage = value; });
    customContextMenuActions.push_back(keepOnlyLastMessageAction);

    useTimestampAction = std::make_shared<QAction>("Use timestamp");
    useTimestampAction->setCheckable(true);
    connect(useTimestampAction.get(), &QAction::toggled, this,
            [=](bool value) { useTimestamp = value; });
    customContextMenuActions.push_back(useTimestampAction);
}
