/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "BaseMavlinkModule.h"

#include <Core/Message/Filter.h>
#include <Core/MessageBroker/MessageBroker.h>
#include <Modules/SkywardHubStrings.h>

int BaseMavlinkModule::ACTIVE_MAVLINK_MODULES = 0;

BaseMavlinkModule::BaseMavlinkModule(MavlinkPort &port, ModuleId id)
    : Module(id), port(port), mavlinkCodec(port, this)
{
    setupUi();

    connect(&mavlinkCodec, &MavlinkCodec::msgReceived, this,
            &BaseMavlinkModule::onMsgReceived);
    connect(&linkQualityTimer, &QTimer::timeout, this,
            &BaseMavlinkModule::onLinkQualityTimerTick);

    subscription = MessageBroker::getInstance().subscribe(
        Filter::fromString(SkywardHubStrings::commandsTopic + "/*"),
        [this](const Message &message, const Filter &filter)
        { onCommandReceived(message); });
}

BaseMavlinkModule::~BaseMavlinkModule()
{
    if (startToggleButton->state())
        stop();
}

XmlObject BaseMavlinkModule::toXmlObject()
{
    XmlObject obj = childToXmlObject();
    obj.addAttribute("log_file", logCheckBox->isChecked() ? 1 : 0);

    return obj;
}

void BaseMavlinkModule::fromXmlObject(const XmlObject &obj)
{
    childFromXmlObject(obj);

    int logFile;
    obj.getAttribute("log_file", logFile);
    logCheckBox->setChecked(logFile != 0);
}

void BaseMavlinkModule::onMsgReceived(const Message &msg)
{
    msgArrived++;
    MessageBroker::getInstance().publish(msg);
}

void BaseMavlinkModule::onLinkQualityTimerTick()
{
    float ratio = (float)msgArrived / linkQualityPeriod * 1000.0;
    msgArrived  = 0;
    rateLabel->setText("Rate: " + QString::number(ratio) + " msg/s");
}

void BaseMavlinkModule::onStartStreamToggled(bool state)
{
    if (state)
        onStartClicked();
    else
        onStopClicked();
}

void BaseMavlinkModule::onStartClicked()
{
    if (open())
    {
        // Keep track of all active and listening mavlink modules
        ACTIVE_MAVLINK_MODULES++;

        disableControls();
        linkQualityTimer.start(linkQualityPeriod);

        // TODO: Allow the checkbox to start and stop the logging
        if (logCheckBox->isChecked())
            mavlinkCodec.openLogFile();

        mavlinkCodec.startReading();
    }
    else
    {
        error(tr("Mavlink"), tr("Unable to open selected port."));
        startToggleButton->setState(false);
    }
}

void BaseMavlinkModule::onStopClicked()
{
    stop();
    port.close();
    enableControls();

    // Keep track of all active and listening mavlink modules
    ACTIVE_MAVLINK_MODULES--;
}

void BaseMavlinkModule::onCommandReceived(const Message &msg)
{
    MavlinkCodec::SysId sysId =
        static_cast<MavlinkCodec::SysId>(sysIdComboBox->currentData().toInt());

    mavlink_message_t mav_msg;
    if (!mavlinkCodec.encodeMessage(msg, sysId, MavlinkCodec::CompIdNone,
                                    mav_msg))
    {
        error(tr("Mavlink"),
              tr("Cannot encode command into a valid mavlink message."));
        return;
    }

    // If we are not active, don't process this message
    if (!startToggleButton->state())
    {
        // Only show this error if there are no other active modules
        if (ACTIVE_MAVLINK_MODULES == 0)
        {
            error(tr("Mavlink"), tr("Cannot send message: port is closed."));
        }

        return;
    }

    mavlinkCodec.send(mav_msg);

    Message confirmationMsg(
        Topic((QString)SkywardHubStrings::logCommandsTopic));

    auto nameField = Field(msg.getTopic().toString().replace(
        SkywardHubStrings::commandsTopic + "/", ""));
    confirmationMsg.setField("name", nameField);

    auto idField = Field((uint64_t)mav_msg.msgid);
    confirmationMsg.setField("message_id", idField);

    auto seqField = Field((uint64_t)mav_msg.seq);
    confirmationMsg.setField("sequence_number", seqField);

    MessageBroker::getInstance().publish(confirmationMsg);
}

void BaseMavlinkModule::onOpenLogFolderClick()
{
    QDir dir(SkywardHubStrings::defaultLogsFolder);
    if (dir.exists())
    {
        QDesktopServices::openUrl(
            QUrl::fromLocalFile(SkywardHubStrings::defaultLogsFolder));
    }
    else
    {
        warning(tr("Mavlink"), tr("The log folder does not exist yet. It will "
                                  "be created when opening the serial port."));
    }
}

void BaseMavlinkModule::enableControls()
{
    logCheckBox->setEnabled(true);
    childEnableControls();
}

void BaseMavlinkModule::disableControls()
{
    logCheckBox->setEnabled(false);
    childDisableControls();
}

void BaseMavlinkModule::stop()
{
    linkQualityTimer.stop();
    mavlinkCodec.stopReading();
}

void BaseMavlinkModule::setupUi()
{
    QHBoxLayout *outerLayout = new QHBoxLayout;
    outerLayout->setContentsMargins(6, 0, 6, 0);

    startToggleButton = std::make_unique<ToggleButton>();
    startToggleButton->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    outerLayout->addWidget(startToggleButton.get());
    connect(startToggleButton.get(), &ToggleButton::toggled, this,
            &BaseMavlinkModule::onStartStreamToggled);

    QLabel *sysIdLabel = new QLabel("Target system:");
    sysIdLabel->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    outerLayout->addWidget(sysIdLabel);

    sysIdComboBox = std::make_unique<QComboBox>();
    sysIdComboBox->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));

    sysIdComboBox->addItem("Main", MavlinkCodec::SysIdMain);
    sysIdComboBox->addItem("Payload", MavlinkCodec::SysIdPayload);
    sysIdComboBox->addItem("Rig", MavlinkCodec::SysIdRig);
    sysIdComboBox->addItem("Gs Receiver", MavlinkCodec::SysIdGS);
    sysIdComboBox->addItem("ARP Receiver", MavlinkCodec::SysIdARP);
    outerLayout->addWidget(sysIdComboBox.get());

    childLayout = std::make_unique<QHBoxLayout>();
    outerLayout->addLayout(childLayout.get());

    outerLayout->addStretch();

    rateLabel = std::make_unique<QLabel>("Rate: 0 msg/s");
    rateLabel->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    outerLayout->addWidget(rateLabel.get());

    logCheckBox = std::make_unique<QCheckBox>("Enable log");
    logCheckBox->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    outerLayout->addWidget(logCheckBox.get());

    QPushButton *openLogFolderButton = new QPushButton("Log folder");
    openLogFolderButton->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    outerLayout->addWidget(openLogFolderButton);
    connect(openLogFolderButton, &QPushButton::clicked, this,
            &BaseMavlinkModule::onOpenLogFolderClick);

    setLayout(outerLayout);
}