/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "OrientationVisualizer.h"

#include <Core/Message/Filter.h>
#include <Core/MessageBroker/MessageBroker.h>

#include <QBoxLayout>
#include <QLabel>
#include <Qt3DExtras/QConeMesh>
#include <Qt3DExtras/QPhongMaterial>
#include <Qt3DExtras/Qt3DWindow>
#include <Qt3DRender/QAttribute>
#include <Qt3DRender/QBuffer>
#include <Qt3DRender/QCamera>
#include <QtMath>

OrientationVisualizer::OrientationVisualizer()
    : Module(ModuleId::ORIENTATION_VISUALIZER)
{

    setupUi();

    updateOrientation(QQuaternion(0.707, 0, 0.707, 0));

    rocketFlightTMSub = MessageBroker::getInstance().subscribe(
        Filter::fromString("Mav/ROCKET_FLIGHT_TM"),
        [this](const Message &message, const Filter &filter)
        {
            if (nullptr == sourceComboBox)
                return;

            if (sourceComboBox->currentText() == "Rocket")
            {
                QQuaternion q(
                    static_cast<float>(message.getField("nas_qw").getDouble()),
                    static_cast<float>(message.getField("nas_qx").getDouble()),
                    static_cast<float>(message.getField("nas_qy").getDouble()),
                    static_cast<float>(message.getField("nas_qz").getDouble()));
                updateOrientation(q);
            }

            if (sourceComboBox->currentText() == "Rocket VN100")
            {
                QQuaternion q(static_cast<float>(
                                  message.getField("vn100_qw").getDouble()),
                              static_cast<float>(
                                  message.getField("vn100_qx").getDouble()),
                              static_cast<float>(
                                  message.getField("vn100_qy").getDouble()),
                              static_cast<float>(
                                  message.getField("vn100_qz").getDouble()));
                updateOrientation(q);
            }
        });

    payloadFlightTMSub = MessageBroker::getInstance().subscribe(
        Filter::fromString("Mav/PAYLOAD_FLIGHT_TM"),
        [this](const Message &message, const Filter &filter)
        {
            if (nullptr == sourceComboBox)
                return;

            if (sourceComboBox->currentText() == "Payload")
            {
                QQuaternion q(
                    static_cast<float>(message.getField("nas_qw").getDouble()),
                    static_cast<float>(message.getField("nas_qx").getDouble()),
                    static_cast<float>(message.getField("nas_qy").getDouble()),
                    static_cast<float>(message.getField("nas_qz").getDouble()));
                updateOrientation(q);
            }
        });
}

OrientationVisualizer::~OrientationVisualizer() {}

void OrientationVisualizer::updateOrientation(QQuaternion q)
{
    float yaw, pitch, roll;

    q.getEulerAngles(&pitch, &yaw, &roll);

    yawLabel->setText(QString::asprintf("Y: %.2lf", yaw));
    pitchLabel->setText(QString::asprintf("P: %.2lf", pitch));
    rollLabel->setText(QString::asprintf("R: %.2lf", roll));

    auto nedDir = q.rotatedVector(QVector3D(1, 0, 0));
    auto stdDir = NedToStdVector(nedDir.x(), nedDir.y(), nedDir.z());

    auto rotation = rotationFromDirections(QVector3D(1, 0, 0), stdDir);

    rocketTransform->setRotation(rotation);
}

void OrientationVisualizer::setupUi()
{
    view                     = std::make_unique<Qt3DExtras::Qt3DWindow>();
    QWidget *windowContainer = QWidget::createWindowContainer(view.get());

    setupScene();

    auto *container = new QVBoxLayout(this);
    container->addWidget(windowContainer);
    container->setStretch(0, 1);
    container->setStretch(1, 0);

    auto lower  = new QHBoxLayout();
    sourceLabel = std::make_unique<QLabel>();
    sourceLabel->setText("Source:");
    sourceLabel->setSizePolicy(
        QSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum));
    sourceComboBox = std::make_unique<QComboBox>();
    sourceComboBox->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    sourceComboBox->addItems({"Rocket", "Rocket VN100", "Payload"});

    lower->addWidget(sourceLabel.get());
    lower->addWidget(sourceComboBox.get());
    lower->addStretch(1);
    addRotationLabels(lower);
    container->addLayout(lower);
}

void OrientationVisualizer::setupScene()
{
    root = std::make_unique<Qt3DCore::QEntity>();
    view->setRootEntity(root.get());

    auto backgroundColor = QColor(QRgb(0x4f4f4f));
    view->defaultFrameGraph()->setClearColor(backgroundColor);

    drawAxes();
    addCamera();
    addLight();
    addGrid();
    addRocket();
}

void OrientationVisualizer::drawAxes()
{
    // NED: North-East-Down
    drawAxis(NedToStdVector(1, 0, 0), QColorConstants::Red);
    drawAxis(NedToStdVector(0, 1, 0), QColorConstants::Green);
    drawAxis(NedToStdVector(0, 0, 1), QColorConstants::Blue);
}

void OrientationVisualizer::drawAxis(QVector3D dir, const QColor &color)
{
    auto length = 15.0f;

    auto rotation = rotationFromDirections(QVector3D(0, 1, 0), dir);

    // Body
    auto cubeEntity = std::make_unique<Qt3DCore::QEntity>(root.get());

    auto cubeMesh = new Qt3DExtras::QCuboidMesh;
    cubeMesh->setXExtent(0.5f);
    cubeMesh->setYExtent(0.5f);
    cubeMesh->setZExtent(0.5f);
    auto cubeTransform = new Qt3DCore::QTransform;
    cubeTransform->setScale3D(QVector3D(1, 1, 1) + dir * length * 2);
    cubeTransform->setTranslation(dir * length / 2);
    auto cubeMaterial = new Qt3DExtras::QGoochMaterial;
    cubeMaterial->setDiffuse(color);
    cubeMaterial->setCool(QColorConstants::Black);
    cubeMaterial->setAlpha(1.0f);

    cubeEntity->addComponent(cubeMesh);
    cubeEntity->addComponent(cubeTransform);
    cubeEntity->addComponent(cubeMaterial);

    // Head
    auto coneEntity = new Qt3DCore::QEntity(root.get());

    auto coneMesh = new Qt3DExtras::QConeMesh;
    coneMesh->setLength(2.0f);
    coneMesh->setBottomRadius(1.0f);
    coneMesh->setSlices(12);
    auto coneTransform = new Qt3DCore::QTransform;
    coneTransform->setTranslation(dir * (length + 1.0f));
    coneTransform->setRotation(rotation);
    auto coneMaterial = new Qt3DExtras::QGoochMaterial;
    coneMaterial->setDiffuse(color);
    coneMaterial->setCool(QColorConstants::Black);
    coneMaterial->setAlpha(1.0f);

    coneEntity->addComponent(coneMesh);
    coneEntity->addComponent(coneTransform);
    coneEntity->addComponent(coneMaterial);
}

void OrientationVisualizer::addLight()
{
    auto *lightEntity = new Qt3DCore::QEntity(root.get());
    auto *light       = new Qt3DRender::QPointLight(lightEntity);
    light->setColor("white");
    light->setIntensity(1);
    auto *lightTransform = new Qt3DCore::QTransform(lightEntity);
    lightTransform->setTranslation(view->camera()->position());

    lightEntity->addComponent(light);
    lightEntity->addComponent(lightTransform);
}

void OrientationVisualizer::addCamera()
{
    camera = view->camera();
    camera->lens()->setOrthographicProjection(
        -PROJECTION_SIZE * 0.5f, PROJECTION_SIZE * 0.5f,
        -PROJECTION_SIZE * 0.5f, PROJECTION_SIZE * 0.5f, 0.1, 100);
    camera->setPosition(QVector3D(30, 30, 30));
    camera->setViewCenter(QVector3D(0, 0, 0));

    // Set camera controls
    auto controller = new Qt3DExtras::QOrbitCameraController(root.get());
    controller->setLinearSpeed(50.0f);
    controller->setLookSpeed(180.0f);
    controller->setEnabled(true);
    controller->setCamera(camera);
}

void OrientationVisualizer::addRocket()
{
    auto rocketEntity = new Qt3DCore::QEntity(root.get());

    auto rocketMesh = new Qt3DRender::QMesh;
    QUrl modelUrl(QStringLiteral("qrc:/assets/models/rocket.obj"));
    rocketMesh->setSource(modelUrl);

    rocketTransform = std::make_unique<Qt3DCore::QTransform>();
    rocketTransform->setScale(3);

    auto rocketWarmColor = QColor(QRgb(0xf2f2f2));
    auto rocketCoolColor = QColor(QRgb(0xbfbfbf));

    auto rocketMaterial = new Qt3DExtras::QGoochMaterial;
    rocketMaterial->setWarm(rocketWarmColor);
    rocketMaterial->setCool(rocketCoolColor);

    rocketEntity->addComponent(rocketMesh);
    rocketEntity->addComponent(rocketTransform.get());
    rocketEntity->addComponent(rocketMaterial);
}

void OrientationVisualizer::addGrid()
{
    QColor gridColor(21, 24, 22);

    for (int i = -GRID_SPAN; i <= GRID_SPAN; i++)
    {
        float xPosition = (float)i * GRID_GAP;

        auto cubeEntity = new Qt3DCore::QEntity(root.get());

        auto cubeMesh = new Qt3DExtras::QCuboidMesh;
        cubeMesh->setXExtent(0.1f);
        cubeMesh->setYExtent(0.1f);
        cubeMesh->setZExtent(0.1f);
        auto cubeTransform = new Qt3DCore::QTransform;
        cubeTransform->setScale3D(
            QVector3D(1, 1, 10 * 2.f * (float)GRID_SPAN * GRID_GAP));
        cubeTransform->setTranslation(QVector3D(xPosition, 0, 0));
        auto cubeMaterial = new Qt3DExtras::QGoochMaterial;
        cubeMaterial->setDiffuse(gridColor);
        cubeMaterial->setWarm(gridColor);
        cubeMaterial->setCool(gridColor);
        cubeMaterial->setSpecular(gridColor);
        cubeMaterial->setShininess(0);
        cubeMaterial->setAlpha(0.3f);

        cubeEntity->addComponent(cubeMesh);
        cubeEntity->addComponent(cubeTransform);
        cubeEntity->addComponent(cubeMaterial);
    }

    for (int j = -GRID_SPAN; j <= GRID_SPAN; j++)
    {
        float yPosition = (float)j * GRID_GAP;

        auto cubeEntity = new Qt3DCore::QEntity(root.get());

        auto cubeMesh = new Qt3DExtras::QCuboidMesh;
        cubeMesh->setXExtent(0.1f);
        cubeMesh->setYExtent(0.1f);
        cubeMesh->setZExtent(0.1f);
        auto cubeTransform = new Qt3DCore::QTransform;
        cubeTransform->setScale3D(
            QVector3D(10 * 2.f * (float)GRID_SPAN * GRID_GAP, 1, 1));
        cubeTransform->setTranslation(QVector3D(0, 0, yPosition));
        auto cubeMaterial = new Qt3DExtras::QGoochMaterial;
        cubeMaterial->setDiffuse(gridColor);
        cubeMaterial->setWarm(gridColor);
        cubeMaterial->setCool(gridColor);
        cubeMaterial->setSpecular(gridColor);
        cubeMaterial->setShininess(0);
        cubeMaterial->setAlpha(0.3f);

        cubeEntity->addComponent(cubeMesh);
        cubeEntity->addComponent(cubeTransform);
        cubeEntity->addComponent(cubeMaterial);
    }
}

void OrientationVisualizer::addRotationLabels(QHBoxLayout *container)
{
    // Visualizer of roll/pitch/yaw

    yawLabel = std::make_unique<QLabel>(this);
    yawLabel->setText("Y: 0.00");
    yawLabel->setAlignment(Qt::AlignCenter);
    yawLabel->setStyleSheet("QLabel { color : green; }");

    pitchLabel = std::make_unique<QLabel>(this);
    pitchLabel->setText("P: 0.00");
    pitchLabel->setAlignment(Qt::AlignCenter);
    pitchLabel->setStyleSheet("QLabel { color : red; }");

    rollLabel = std::make_unique<QLabel>(this);
    rollLabel->setText("R: 0.00");
    rollLabel->setAlignment(Qt::AlignCenter);
    rollLabel->setStyleSheet("QLabel { color : blue; }");
    this->setStyleSheet("QLabel {font-weight: bold;}");

    container->addWidget(yawLabel.get());
    container->addWidget(pitchLabel.get());
    container->addWidget(rollLabel.get());
}

QQuaternion OrientationVisualizer::rotationFromDirections(QVector3D start,
                                                          QVector3D end)
{
    auto axis = QVector3D::crossProduct(start, end);
    if (axis.length() == 0)
        axis = {1, 0, 0};
    auto radAngle = std::acos(QVector3D::dotProduct(start, end));
    auto degAngle = qRadiansToDegrees(radAngle);

    return QQuaternion::fromAxisAndAngle(axis, degAngle);
}

QVector3D OrientationVisualizer::NedToStdVector(float x, float y, float z)
{
    return {x, -z, y};
}

QVector3D OrientationVisualizer::stdToNedVector(float x, float y, float z)
{
    return {x, z, -y};
}

void OrientationVisualizer::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);

    int width         = this->width();
    int height        = this->height();
    float aspectRatio = (float)(height) / (float)(width);

    // Update camera aspect ratio to avoid stretching
    // camera.setAspectRatio does not work
    // So we set left/right/top/bottom manually
    if (height > width)
    {
        camera->setLeft((float)PROJECTION_SIZE * -0.5f);
        camera->setRight((float)PROJECTION_SIZE * 0.5f);
        camera->setBottom((float)PROJECTION_SIZE * -0.5f * aspectRatio);
        camera->setTop((float)PROJECTION_SIZE * 0.5f * aspectRatio);
    }
    else
    {
        camera->setLeft((float)PROJECTION_SIZE * -0.5f / aspectRatio);
        camera->setRight((float)PROJECTION_SIZE * 0.5f / aspectRatio);
        camera->setBottom((float)PROJECTION_SIZE * -0.5f);
        camera->setTop((float)PROJECTION_SIZE * 0.5f);
    }
}
