/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "PayloadStateViewer.h"

#include <Components/FilterSelector/FilterSelector.h>
#include <Core/MessageBroker/MessageBroker.h>

PayloadStateViewerModule::PayloadStateViewerModule()
    : Module(ModuleId::PAYLOAD_STATE_VIEWER)
{
    setupUi();
    customContextMenuActionSetup();
}

PayloadStateViewerModule::~PayloadStateViewerModule() {}

XmlObject PayloadStateViewerModule::toXmlObject()
{
    XmlObject obj = Module::toXmlObject();

    obj.addAttribute("filter", filter.toString());

    return obj;
}

void PayloadStateViewerModule::fromXmlObject(const XmlObject& xmlObject)
{
    auto filter = Filter::fromString(xmlObject.getAttribute("filter"));
    setFilter(filter);
}

void PayloadStateViewerModule::setupUi()
{
    outerLayout = std::make_unique<QHBoxLayout>();
    outerLayout->setContentsMargins(0, 0, 0, 0);
    outerLayout->setSpacing(0);

    for (auto labelText : PayloadStatesList::statesLabels)
    {
        QLabel* label = new QLabel;
        label->setText(labelText);
        label->setAlignment(Qt::AlignCenter);
        label->setContentsMargins(4, 4, 4, 4);
        outerLayout->addWidget(label);
    }

    setLayout(outerLayout.get());
}

void PayloadStateViewerModule::customContextMenuActionSetup()
{
    std::shared_ptr<QAction> action =
        std::make_shared<QAction>("Choose topic and field");
    connect(action.get(), &QAction::triggered, this,
            &PayloadStateViewerModule::onConfigureClicked);

    customContextMenuActions.push_back(action);
}

void PayloadStateViewerModule::onConfigureClicked()
{
    FilterSelector::selectFilter(
        filter, [this](const Filter& newFilter) { setFilter(newFilter); });
}

void PayloadStateViewerModule::setFilter(const Filter& newFilter)
{
    subscription = MessageBroker::getInstance().subscribe(
        newFilter, [this](const Message& message, const Filter& filter)
        { onMsgReceived(message); });
    filter = newFilter;
}

void PayloadStateViewerModule::onMsgReceived(const Message& msg)
{
    Field field = msg.getField("fmm_state");

    PayloadStatesList::State state =
        static_cast<PayloadStatesList::State>(field.getUnsignedInteger());

    // Skip invalid and macro states
    if (state == PayloadStatesList::State::INVALID)
        return;

    if (state != currentState)
    {
        currentState = static_cast<PayloadStatesList::State>(state);

        QString baseStyle =
            "border-bottom-width:1px;border-left-width:1px;border-top-width:"
            "1px;border-radius:0;";
        QString currentStateStyle =
            "background-color:yellow; color:black;" + baseStyle;
        QString completedStyle = "background-color:green;" + baseStyle;
        QString errorStyle     = "background-color:red;" + baseStyle;

        int index  = PayloadStatesList::statesIndexes[currentState];
        auto label = outerLayout->itemAt(index)->widget();

        if (currentState == PayloadStatesList::State::INIT_ERROR)
            label->setStyleSheet(errorStyle);
        else
            label->setStyleSheet(currentStateStyle);

        // Set as completed every state before the current
        for (int i = index - 1; i >= 0; i--)
            outerLayout->itemAt(i)->widget()->setStyleSheet(completedStyle);

        // Reset every state after the current
        for (int i = index + 1; i < PayloadStatesList::statesLabels.count();
             i++)
            outerLayout->itemAt(i)->widget()->setStyleSheet(baseStyle);

        auto tmp =
            outerLayout->itemAt(PayloadStatesList::statesLabels.count() - 1)
                ->widget()
                ->styleSheet();
        outerLayout->itemAt(PayloadStatesList::statesLabels.count() - 1)
            ->widget()
            ->setStyleSheet(tmp + "border-right-width:1px;");
    }
}
