/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "Splitter.h"

#include <Modules/ModulesList.h>

#include <QBoxLayout>

Splitter::Splitter(Qt::Orientation orientation, ModuleId id) : Module(id)
{
    setupUi(orientation);
    customContextMenuActionSetup();
}

XmlObject Splitter::toXmlObject()
{
    auto obj = Module::toXmlObject();

    QString sizes;
    for (int i = 0; i < splitter->sizes().size(); i++)
    {
        sizes += QString::number(splitter->sizes()[i]);

        if (i < splitter->sizes().size() - 1)
            sizes += ",";
    }

    obj.addAttribute("sizes", sizes);

    for (int i = 0; i < splitter->count(); i++)
    {
        Module* module = qobject_cast<Module*>(splitter->widget(i));
        obj.addChild(module->toXmlObject());
    }

    return obj;
}

void Splitter::fromXmlObject(const XmlObject& obj)
{
    QList<int> intSizes;
    for (QString size : obj.getAttribute("sizes").split(","))
        intSizes.append(size.toInt());

    for (int i = 0; i < obj.childCount(); i++)
    {
        XmlObject xmlChild = obj.childAt(i);
        std::shared_ptr<Module> module =
            ModulesList::getInstance().instantiateModule(
                xmlChild.getObjectName());

        if (module)
        {
            module->fromXmlObject(xmlChild);
            addModule(module, i);
        }
        else
        {
            // TODO: what should we do?
        }
    }

    splitter->setSizes(intSizes);
}

void Splitter::addModule(std::shared_ptr<Module> module)
{
    addModule(module, splitter->count());
}

void Splitter::addModule(std::shared_ptr<Module> module, int position)
{
    connect(module.get(), &Module::replaceMe, this, &Splitter::replaceChild);
    connect(module.get(), &Module::closeMe, this, &Splitter::closeChild);

    splitter->insertWidget(position, qobject_cast<QWidget*>(module.get()));
    children[position] = module;
}

void Splitter::replaceChild(Module* oldModule,
                            std::shared_ptr<Module> newModule)
{
    // Check if the modules are valid
    if (!oldModule || !newModule)
        return;

    // Check if we are trying to replace a module with itself
    if (oldModule == newModule.get())
        return;

    auto index = splitter->indexOf(oldModule);

    // Check if the old module is a child
    if (index == -1)
        return;

    addModule(newModule, index);
}

void Splitter::closeChild(Module* module)
{
    // Check if the given module is a child
    if (splitter->indexOf(module) == -1)
        return;

    // If there is only one child, then we have to delete the whole splitter
    if (splitter->count() == 1)
    {
        emit closeMe(this);
        return;
    }

    // If there is only two children, then the remaining one will replace the
    // whole splitter
    if (splitter->count() == 2)
    {
        std::shared_ptr<Module> currentChild = children[0];

        // REmove the children parent. This is essential otherwise when the
        // splitter will be deleted, also the child will be
        currentChild->setParent(nullptr);

        // Replace the splitter with the only left children
        emit replaceMe(this, currentChild);
    }
}

Qt::Orientation Splitter::getOrientation() const
{
    return splitter->orientation();
}

void Splitter::setOrientation(Qt::Orientation orientation)
{
    splitter->setOrientation(orientation);
}

void Splitter::swapOrientation()
{
    if (getOrientation() == Qt::Horizontal)
    {
        id = ModuleId::VERTICAL_SPLITTER;
        splitter->setOrientation(Qt::Vertical);
    }
    else
    {
        id = ModuleId::HORIZONTAL_SPLITTER;
        splitter->setOrientation(Qt::Horizontal);
    }
}

void Splitter::customContextMenuActionSetup()
{
    std::shared_ptr<QAction> split = std::make_shared<QAction>("Split");
    connect(split.get(), &QAction::triggered, this,
            [this]() {
                addModule(ModulesList::getInstance().instantiateModule(
                    ModuleId::EMPTY));
            });
    customContextMenuActions.push_back(split);

    std::shared_ptr<QAction> changeOrientation =
        std::make_shared<QAction>("Swap Orientation");
    connect(changeOrientation.get(), &QAction::triggered, this,
            [this]() { this->swapOrientation(); });
    customContextMenuActions.push_back(changeOrientation);
}

void Splitter::setupUi(Qt::Orientation orientation)
{
    QVBoxLayout* outerLayout = new QVBoxLayout;
    outerLayout->setContentsMargins(0, 0, 0, 0);

    splitter = std::make_unique<QSplitter>();
    splitter->setOrientation(orientation);
    splitter->setChildrenCollapsible(false);
    splitter->setHandleWidth(0);
    outerLayout->addWidget(splitter.get());

    // By default add two default modules
    addModule(ModulesList::getInstance().instantiateModule(ModuleId::EMPTY));
    addModule(ModulesList::getInstance().instantiateModule(ModuleId::EMPTY));

    setLayout(outerLayout);
}
