/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "TabsModule.h"

#include <Modules/ModulesList.h>

#include <QBoxLayout>
#include <QDebug>
#include <QLabel>

TabsModule::TabsModule() : Module(ModuleId::TABS)
{
    setupUi();
    customContextMenuActionSetup();

    QObject::connect(tabNames.get(), &QListWidget::currentRowChanged,
                     tabContents.get(), &QStackedWidget::setCurrentIndex);
    QObject::connect(tabNames.get(), &QListWidget::itemDoubleClicked, this,
                     &TabsModule::onDoubleClickedItem);

    // Automatically add first tab
    addTab(generateUniqueTabName(),
           ModulesList::getInstance().instantiateModule(ModuleId::EMPTY));
}

XmlObject TabsModule::toXmlObject()
{
    XmlObject obj = Module::toXmlObject();

    for (int i = 0; i != tabContents->count(); i++)
    {
        XmlObject tab("tab");
        tab.addAttribute("name", tabNames->item(i)->text());
        if (auto module = qobject_cast<Module*>(tabContents->widget(i)))
        {
            tab.addChild(module->toXmlObject());
        }
        else
        {
            qCritical() << "Unable to save tab " << tabNames->item(i)->text()
                        << " due to failed when casting widget to Module*";
            continue;  // Skip this tab
        }

        obj.addChild(tab);
    }

    return obj;
}

void TabsModule::fromXmlObject(const XmlObject& xmlObject)
{
    // Discard previous content
    tabNames->clear();
    modules.clear();

    for (int i = 0; i < xmlObject.childCount(); ++i)
    {
        XmlObject tabNode = xmlObject.childAt(i);
        if (tabNode.getObjectName() != "tab" || tabNode.childCount() == 0)
            continue;

        XmlObject xmlChild = tabNode.childAt(0);

        std::shared_ptr<Module> module =
            ModulesList::getInstance().instantiateModule(
                xmlChild.getObjectName());
        if (module != nullptr)
        {
            module->fromXmlObject(xmlChild);
            addTab(tabNode.getAttribute("name", generateUniqueTabName()),
                   module);
        }
        else
        {
            qCritical() << "Failed to instantiate module"
                        << tabNode.getAttribute("name",
                                                generateUniqueTabName());
        }
    }

    // If the tabs are left empty, add a single tab with an empty module
    if (tabNames->count() == 0)
        addTab(generateUniqueTabName(),
               ModulesList::getInstance().instantiateModule(ModuleId::EMPTY));

    tabNames->setCurrentItem(tabNames->item(0));
}

void TabsModule::addTab(const QString& tabName, std::shared_ptr<Module> module)
{
    connect(module.get(), &Module::replaceMe, this, &TabsModule::replaceTab);
    connect(module.get(), &Module::closeMe, this, &TabsModule::closeTab);

    auto* listItem = new QListWidgetItem();
    listItem->setText(tabName);
    listItem->setTextAlignment(Qt::AlignCenter);
    tabNames->addItem(listItem);

    tabContents->addWidget(module.get());

    // We set the correct tabName after the widget is added
    tabNames->setCurrentItem(listItem);
    modules.push_back(module);
}

QString TabsModule::generateUniqueTabName()
{
    return QString("Tab %1").arg(tabNames->count());
}

void TabsModule::replaceTab(Module* const oldModule,
                            std::shared_ptr<Module> newModule)
{
    // Check if the modules are valid
    if (!oldModule || !newModule)
        return;

    // Check if we are trying to replace a module with itself
    if (oldModule == newModule.get())
        return;

    // Get the index of the old module
    int index = tabContents->indexOf(oldModule);
    if (index == -1)
        return;

    // Replace the module
    tabContents->insertWidget(index, newModule.get());
    tabContents->setCurrentWidget(newModule.get());
    tabContents->removeWidget(oldModule);
    tabContents->update();

    connect(newModule.get(), &Module::replaceMe, this, &TabsModule::replaceTab);
    connect(newModule.get(), &Module::closeMe, this, &TabsModule::closeTab);
    modules[index] = newModule;
}

void TabsModule::closeTab(Module* module)
{
    int index = tabContents->indexOf(module);

    // Check if the given module is a child
    if (index == -1)
        return;

    // If there is only one child, then close the entire tabs module
    if (tabContents->count() == 1)
    {
        emit closeMe(this);
        return;
    }

    tabContents->removeWidget(module);

    modules.erase(std::remove_if(modules.begin(), modules.end(),
                                 [module](std::shared_ptr<Module>& value)
                                 { return value.get() == module; }));

    delete tabNames->item(index);
    tabNames->setCurrentItem(tabNames->item(index == 0 ? 0 : index - 1));
}

void TabsModule::onDoubleClickedItem(QListWidgetItem* item)
{
    item->setFlags(item->flags() | Qt::ItemIsEditable);
    tabNames->setCurrentItem(item);
    tabNames->editItem(item);
}

void TabsModule::customContextMenuActionSetup()
{
    std::shared_ptr<QAction> newTab = std::make_shared<QAction>("New Tab");
    connect(newTab.get(), &QAction::triggered, this,
            &TabsModule::onMenuNewTabClick);
    customContextMenuActions.push_back(newTab);

    std::shared_ptr<QAction> deleteTab =
        std::make_shared<QAction>("Delete Tab");
    connect(deleteTab.get(), &QAction::triggered, this,
            &TabsModule::onMenuDeleteTabClick);
    customContextMenuActions.push_back(deleteTab);

    std::shared_ptr<QAction> renameTab =
        std::make_shared<QAction>("Rename Tab");
    connect(renameTab.get(), &QAction::triggered, this,
            &TabsModule::onMenuRenameTabClick);
    customContextMenuActions.push_back(renameTab);
}

void TabsModule::onMenuNewTabClick()
{
    addTab(generateUniqueTabName(),
           ModulesList::getInstance().instantiateModule(ModuleId::EMPTY));
}

void TabsModule::onMenuDeleteTabClick()
{
    if (auto module = qobject_cast<Module*>(tabContents->currentWidget()))
        closeTab(module);
    else
        qCritical() << "Unable to close tab  due to failed when casting widget "
                       "to Module*";
}

void TabsModule::onMenuRenameTabClick()
{
    auto* item = tabNames->currentItem();
    item->setFlags(item->flags() | Qt::ItemIsEditable);
    tabNames->editItem(item);
}

void TabsModule::setupUi()
{
    // Top component (top bar)
    tabNames = std::make_unique<QListWidget>();
    tabNames->setFlow(QListView::Flow::LeftToRight);
    tabNames->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
    tabNames->setSizeAdjustPolicy(QAbstractScrollArea::AdjustToContents);
    tabNames->setSelectionMode(QAbstractItemView::SingleSelection);
    tabNames->setMaximumHeight(40);

    // Bottom component
    tabContents = std::make_unique<QStackedWidget>();
    tabContents->setObjectName("TabsStackedWidget");

    // Putting all together
    QBoxLayout* outerLayout = new QBoxLayout(QBoxLayout::TopToBottom, this);
    outerLayout->setContentsMargins(0, 0, 0, 0);
    outerLayout->setSpacing(0);
    outerLayout->addWidget(tabNames.get());
    outerLayout->addWidget(tabContents.get());
}
