/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "ValvesViewer.h"

#include <Components/FilterSelector/FilterSelector.h>
#include <Core/MessageBroker/MessageBroker.h>

ValvesViewer::ValvesViewer() : Module(ModuleId::VALVES_VIEWER)
{
    setupUi();
    customContextMenuActionSetup();
}

ValvesViewer::~ValvesViewer() {}

XmlObject ValvesViewer::toXmlObject()
{
    XmlObject obj = Module::toXmlObject();

    obj.addAttribute("filter", filter.toString());

    return obj;
}

void ValvesViewer::fromXmlObject(const XmlObject& xmlObject)
{
    auto filter = Filter::fromString(xmlObject.getAttribute("filter"));
    setFilter(filter);
}

void ValvesViewer::setupUi()
{
    outerLayout = std::make_unique<QHBoxLayout>();
    outerLayout->setContentsMargins(0, 0, 0, 0);
    outerLayout->setSpacing(0);

    for (auto labelText : ValvesList::valvesLabels)
    {
        QLabel* label = new QLabel;
        label->setText(labelText);
        label->setAlignment(Qt::AlignCenter);
        label->setContentsMargins(4, 4, 4, 4);
        outerLayout->addWidget(label);
    }

    setLayout(outerLayout.get());
}

void ValvesViewer::customContextMenuActionSetup()
{
    std::shared_ptr<QAction> action =
        std::make_shared<QAction>("Choose topic and field");
    connect(action.get(), &QAction::triggered, this,
            &ValvesViewer::onConfigureClicked);

    customContextMenuActions.push_back(action);
}

void ValvesViewer::onConfigureClicked()
{
    FilterSelector::selectFilter(
        filter, [this](const Filter& newFilter) { setFilter(newFilter); });
}

void ValvesViewer::setFilter(const Filter& newFilter)
{
    subscription = MessageBroker::getInstance().subscribe(
        newFilter, [this](const Message& message, const Filter& filter)
        { onMsgReceived(message); });
    filter = newFilter;
}

void ValvesViewer::onMsgReceived(const Message& msg)
{
    // Couple the valves with their own state
    QMap<ValvesList::Valve, uint64_t> map{
        {ValvesList::Valve::FILLING,
         msg.getField("filling_valve_state").getUnsignedInteger()},
        {ValvesList::Valve::MAIN,
         msg.getField("main_valve_state").getUnsignedInteger()},
        {ValvesList::Valve::RELEASE,
         msg.getField("release_valve_state").getUnsignedInteger()},
        {ValvesList::Valve::VENTING,
         msg.getField("venting_valve_state").getUnsignedInteger()},
        {ValvesList::Valve::NITROGEN,
         msg.getField("nitrogen_valve_state").getUnsignedInteger()},
        {ValvesList::Valve::TARS,
         msg.getField("tars_state").getUnsignedInteger()}};

    // Define the "active" and the "not active" look
    QString baseStyle =
        "border-bottom-width:1px;border-left-width:1px;border-top-width:"
        "1px;border-radius:0;";
    QString completedStyle = "background-color:green;" + baseStyle;
    QString errorStyle     = "background-color:red;" + baseStyle;

    // Iterate over the valves to assign them the correct look
    for (auto i = map.cbegin(), end = map.cend(); i != end; i++)
    {
        if (i.value() == 1)
        {
            // The valve is opened
            auto label =
                outerLayout->itemAt(static_cast<int>(i.key()))->widget();
            label->setStyleSheet(completedStyle);
        }
        else
        {
            // The valve is closed
            auto label =
                outerLayout->itemAt(static_cast<int>(i.key()))->widget();
            label->setStyleSheet(errorStyle);
        }
    }
}
